//
// sfive
//
// sfive - spreadspace streaming statistics suite is a generic
// statistic collection tool for streaming server infrastuctures.
// The system collects and stores meta data like number of views
// and throughput from a number of streaming servers and stores
// it in a global data store.
// The data acquisition is designed to be generic and extensible in
// order to support different streaming software.
// sfive also contains tools and applications to filter and visualize
// live and recorded data.
//
//
// Copyright (C) 2014-2018 Christian Pointner <equinox@spreadspace.org>
//                         Markus Grüneis <gimpf@gimpf.org>
//
// This file is part of sfive.
//
// sfive is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 3
// as published by the Free Software Foundation.
//
// sfive is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with sfive. If not, see <http://www.gnu.org/licenses/>.
//

package main

import (
	"encoding/json"
	"errors"
	"fmt"
	"net/http"
	"os"
	"strconv"
	"strings"
	"time"

	"github.com/spreadspace/tlsconfig"
)

//*********************************************
//********** HTTP/HTTPS Proxy Configuration

//********** ProtocolType

type ProtocolType string

const (
	HTTPAndHTTPS ProtocolType = "http+https"
	HTTPOnly     ProtocolType = "http"
	HTTPSOnly    ProtocolType = "https"
)

func (p ProtocolType) String() string {
	return string(p)
}

func (p *ProtocolType) fromString(str string) (err error) {
	switch strings.ToLower(str) {
	case string(HTTPAndHTTPS):
		fallthrough
	case string(HTTPOnly):
		fallthrough
	case string(HTTPSOnly):
		*p = ProtocolType(strings.ToLower(str))
	default:
		return fmt.Errorf("invalid protocol: '"+str+"', must be one of '%s', '%s' or '%s'", HTTPOnly, HTTPSOnly, HTTPAndHTTPS)
	}
	return
}

func (p ProtocolType) MarshalText() (data []byte, err error) {
	data = []byte(p.String())
	return
}

func (p *ProtocolType) UnmarshalText(data []byte) (err error) {
	return p.fromString(string(data))
}

//********** RedirectCode

type RedirectCode int

func (r RedirectCode) String() string {
	return strconv.Itoa(int(r))
}

func (r *RedirectCode) fromString(str string) error {
	code, err := strconv.Atoi(str)
	if err != nil {
		return err
	}
	switch code {
	case 0:
		fallthrough
	case http.StatusMovedPermanently:
		fallthrough
	case http.StatusFound:
		fallthrough
	case http.StatusSeeOther:
		fallthrough
	case http.StatusTemporaryRedirect:
		*r = RedirectCode(code)
		return nil
	}
	return fmt.Errorf("invalid redirect code: '"+str+"', must be one of %d, %d, %d or %d",
		http.StatusMovedPermanently, http.StatusFound, http.StatusSeeOther, http.StatusTemporaryRedirect)
}

func (r RedirectCode) MarshalJSON() (data []byte, err error) {
	data = []byte(r.String())
	return
}

func (r *RedirectCode) UnmarshalJSON(data []byte) (err error) {
	return r.fromString(string(data))
}

//********** Operation

type Operation string

const (
	OpAdd  Operation = "add"
	OpDel  Operation = "del"
	OpSet  Operation = "set"
	OpTime Operation = "time"
)

func (o Operation) String() string {
	return string(o)
}

func (o *Operation) fromString(str string) (err error) {
	switch strings.ToLower(str) {
	case string(OpAdd):
		fallthrough
	case string(OpDel):
		fallthrough
	case string(OpSet):
		fallthrough
	case string(OpTime):
		*o = Operation(strings.ToLower(str))
	default:
		return fmt.Errorf("invalid operation: '"+str+"', must be one of '%s', '%s', '%s' or '%s'", OpAdd, OpDel, OpSet, OpTime)
	}
	return
}

func (o Operation) MarshalText() (data []byte, err error) {
	data = []byte(o.String())
	return
}

func (o *Operation) UnmarshalText(data []byte) (err error) {
	return o.fromString(string(data))
}

//********** HeaderOperation

type HeaderOperation struct {
	Operation     Operation     `json:"op"`
	Header        string        `json:"header"`
	Value         string        `json:"value,omitempty"`
	ValueDuration time.Duration `json:"-"`
}

func (h *HeaderOperation) Parse() (err error) {
	switch h.Operation {
	case OpTime:
		if h.ValueDuration, err = time.ParseDuration(h.Value); err != nil {
			return err
		}
	}
	return nil
}

//*********************************************
//********** SFive Statistics

//********** SFiveDuration

type SFiveDuration int64

func (d *SFiveDuration) UnmarshalText(data []byte) error {
	v, err := time.ParseDuration(string(data))
	if err != nil {
		return err
	}
	if v < 1*time.Second {
		return errors.New("duration must be >= 1s")
	}
	*d = SFiveDuration(v / time.Millisecond)
	return nil
}

//********** SFiveConf

type SFiveConf struct {
	Sock     string        `json:"socket"`
	Hostname string        `json:"hostname"`
	Tags     []string      `json:"tags"`
	Duration SFiveDuration `json:"duration"`
	Matches  []struct {
		Format  string `json:"format"`
		Streams string `json:"streams"`
	} `json:"matches"`
}

//*********************************************
//********** Backend

type Backend struct {
	ConnectAddr    string            `json:"connect"`
	RequestHeader  []HeaderOperation `json:"request_header"`
	ResponseHeader []HeaderOperation `json:"response_header"`
}

//*********************************************
//********** Configuration

type Config struct {
	ListenAddr     string              `json:"listen"`
	Protocol       ProtocolType        `json:"protocol"`
	Redirect2HTTPS RedirectCode        `json:"redirect2https"`
	TLS            tlsconfig.TLSConfig `json:"tls"`

	Backend
	Backends map[string]Backend `json:"backends"`

	SFive SFiveConf `json:"sfive"`
}

func readConfig(configfile string) (conf *Config, err error) {
	s5l.Printf("reading config from: %s", configfile)

	var f *os.File
	if f, err = os.Open(configfile); err != nil {
		return
	}
	defer f.Close()

	conf = &Config{}
	conf.SFive.Duration = 10000
	if err = json.NewDecoder(f).Decode(conf); err != nil {
		return
	}

	for idx, _ := range conf.RequestHeader {
		if err = conf.RequestHeader[idx].Parse(); err != nil {
			return
		}
	}
	for idx, _ := range conf.ResponseHeader {
		if err = conf.ResponseHeader[idx].Parse(); err != nil {
			return
		}
	}

	if conf.Protocol == HTTPAndHTTPS || conf.Protocol == HTTPSOnly {
		if conf.TLS.CertFile == "" || conf.TLS.KeyFile == "" {
			return nil, errors.New("HTTPs is enabled but no certificate and/or key file is supplied")
		}
	}

	for location, backend := range conf.Backends {
		tmp := backend
		if backend.ConnectAddr == "" {
			tmp.ConnectAddr = conf.ConnectAddr
		}
		if len(backend.RequestHeader) == 0 {
			tmp.RequestHeader = conf.RequestHeader
		}
		if len(backend.ResponseHeader) == 0 {
			tmp.ResponseHeader = conf.ResponseHeader
		}
		conf.Backends[location] = tmp
	}

	return
}
