//
// sfive
//
// sfive - spreadspace streaming statistics suite is a generic
// statistic collection tool for streaming server infrastuctures.
// The system collects and stores meta data like number of views
// and throughput from a number of streaming servers and stores
// it in a global data store.
// The data acquisition is designed to be generic and extensible in
// order to support different streaming software.
// sfive also contains tools and applications to filter and visualize
// live and recorded data.
//
//
// Copyright (C) 2014-2018 Christian Pointner <equinox@spreadspace.org>
//                         Markus Grüneis <gimpf@gimpf.org>
//
// This file is part of sfive.
//
// sfive is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 3
// as published by the Free Software Foundation.
//
// sfive is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with sfive. If not, see <http://www.gnu.org/licenses/>.
//

package main

import (
	"crypto/tls"
	"fmt"
	"net"
	"net/http"
	"net/http/httputil"
	"net/url"
	"strings"
	"time"

	"github.com/soheilhy/cmux"
)

func generateTime(d time.Duration) string {
	return time.Now().UTC().Add(d).Format(time.RFC1123)
}

type s5proxyResponseWriter struct {
	wrapped    http.ResponseWriter
	respHeader []HeaderOperation
	StatsCh    chan<- *Client
	IP         string
	Port       int
	UserAgent  string
}

func (r s5proxyResponseWriter) Header() http.Header {
	return r.wrapped.Header()
}

func (r s5proxyResponseWriter) Write(data []byte) (int, error) {
	sent, err := r.wrapped.Write(data)
	if r.StatsCh != nil {
		stats := &Client{IP: r.IP, Port: uint(r.Port), UserAgent: r.UserAgent, BytesSent: uint(sent)}
		select {
		case r.StatsCh <- stats:
		default:
		}
	}
	return sent, err
}

func (r s5proxyResponseWriter) WriteHeader(status int) {
	for _, h := range r.respHeader {
		switch h.Operation {
		case OpAdd:
			r.wrapped.Header().Add(h.Header, h.Value)
		case OpDel:
			r.wrapped.Header().Del(h.Header)
		case OpSet:
			r.wrapped.Header().Set(h.Header, h.Value)
		case OpTime:
			r.wrapped.Header().Set(h.Header, generateTime(h.ValueDuration))
		}
	}
	r.wrapped.WriteHeader(status)
}

func proxyHandler(p *httputil.ReverseProxy, s *Stats, respHeader []HeaderOperation, w http.ResponseWriter, r *http.Request) {
	pw := s5proxyResponseWriter{wrapped: w, respHeader: respHeader}
	if s != nil {
		if ip, port, err := net.SplitHostPort(r.RemoteAddr); err != nil {
			s5l.Printf("PROXY: client '%s' error: invalid address/port info: %v, no statistics will be gathered", r.RemoteAddr, err)
		} else {
			pw.IP = ip
			if pw.Port, err = net.LookupPort("tcp", port); err != nil {
				s5l.Printf("PROXY: client '%s' warning: can't parse port or service name: %v", r.RemoteAddr, err)
			}
			pw.UserAgent = r.UserAgent()
			pw.StatsCh = s.GetUpdateChannel(r.URL.String())
		}
	}

	p.ServeHTTP(pw, r)
}

type Proxy struct {
	conf  *Config
	stats *Stats
	mux   *http.ServeMux
	srv   *http.Server
}

func (p *Proxy) installProxyHandler(location, connectAddr string, reqHeader, respHeader []HeaderOperation) error {
	remote, err := url.Parse(connectAddr)
	if err != nil {
		return err
	}
	s5l.Printf("PROXY: forwarding traffic for %s to '%s'", location, remote.String())

	proxy := httputil.NewSingleHostReverseProxy(remote)
	origDir := proxy.Director
	proxy.Director = func(req *http.Request) {
		origDir(req)
		for _, h := range reqHeader {
			switch h.Operation {
			case OpAdd:
				req.Header.Add(h.Header, h.Value)
			case OpDel:
				req.Header.Del(h.Header)
			case OpSet:
				req.Header.Set(h.Header, h.Value)
			case OpTime:
				req.Header.Set(h.Header, generateTime(h.ValueDuration))
			}
		}
	}

	p.mux.HandleFunc(location, func(w http.ResponseWriter, r *http.Request) {
		proxyHandler(proxy, p.stats, respHeader, w, r)
	})
	return nil
}

func NewProxy(conf *Config, stats *Stats) (p *Proxy, err error) {
	p = &Proxy{conf: conf, stats: stats}
	p.mux = http.NewServeMux()

	for location, backend := range conf.Backends {
		p.installProxyHandler(location, backend.ConnectAddr, backend.RequestHeader, backend.ResponseHeader)
	}
	p.installProxyHandler("/", conf.ConnectAddr, conf.RequestHeader, conf.ResponseHeader)
	p.srv = &http.Server{Handler: p.mux}
	return
}

type httpsRedirectHandler struct {
	code int
}

func (h *httpsRedirectHandler) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	uri := *r.URL
	uri.Scheme = "https"
	uri.Host = r.Host
	http.Redirect(w, r, uri.String(), h.code)
}

func (p *Proxy) RunHTTPRedirector(l net.Listener) error {
	mux := http.NewServeMux()
	mux.Handle("/", &httpsRedirectHandler{int(p.conf.Redirect2HTTPS)})

	srv := &http.Server{Handler: mux}
	return srv.Serve(l)
}

func (p *Proxy) RunHTTP(l net.Listener) error {
	return p.srv.Serve(l)
}

func (p *Proxy) RunHTTPS(l net.Listener) error {
	cfg, err := p.conf.TLS.ToGoTLSConfig()
	if err != nil {
		return err
	}

	tlsL := tls.NewListener(l, cfg)
	return p.srv.Serve(tlsL)
}

func (p *Proxy) Run() error {
	l, err := net.Listen("tcp", p.conf.ListenAddr)
	if err != nil {
		return err
	}

	s5l.Printf("PROXY: listening on '%s' using protocol: %s", l.Addr(), p.conf.Protocol)
	switch p.conf.Protocol {
	case HTTPAndHTTPS:
		m := cmux.New(l)

		httpL := m.Match(cmux.HTTP1Fast())
		if p.conf.Redirect2HTTPS > 0 {
			s5l.Printf("PROXY: will redirect any traffic from http to https using status code %v", p.conf.Redirect2HTTPS)
			go p.RunHTTPRedirector(httpL)
		} else {
			go p.RunHTTP(httpL)
		}

		httpsL := m.Match(cmux.Any())
		go p.RunHTTPS(httpsL)

		if err := m.Serve(); !strings.Contains(err.Error(), "use of closed network connection") { // TODO: is this really the best way to do this?
			return err
		}
		return nil
	case HTTPOnly:
		return p.RunHTTP(l)
	case HTTPSOnly:
		return p.RunHTTPS(l)
	}
	return fmt.Errorf("PROXY: unsupported protocol: '%s'", p.conf.Protocol)
}
