//
// sfive
//
// sfive - spreadspace streaming statistics suite is a generic
// statistic collection tool for streaming server infrastuctures.
// The system collects and stores meta data like number of views
// and throughput from a number of streaming servers and stores
// it in a global data store.
// The data acquisition is designed to be generic and extensible in
// order to support different streaming software.
// sfive also contains tools and applications to filter and visualize
// live and recorded data.
//
//
// Copyright (C) 2014-2017 Christian Pointner <equinox@spreadspace.org>
//                         Markus Grüneis <gimpf@gimpf.org>
//
// This file is part of sfive.
//
// sfive is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 3
// as published by the Free Software Foundation.
//
// sfive is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with sfive. If not, see <http://www.gnu.org/licenses/>.
//

package sfive

import (
	"encoding/json"
	"fmt"
	"io"
)

const (
	ProtocolVersion = 2
)

//
// Decoder
//

type Decoder interface {
	Decode() (*UpdateFull, error)
	Slug() string
}

// stateless protocol interfaces
type StatelessDecoder struct {
	dec *json.Decoder
}

func NewStatelessDecoder(r io.Reader) Decoder {
	return &StatelessDecoder{json.NewDecoder(r)}
}

func (sd *StatelessDecoder) Slug() string {
	return ""
}

func (sd *StatelessDecoder) Decode() (uf *UpdateFull, err error) {
	uf = &UpdateFull{}
	if err = sd.dec.Decode(uf); err != nil {
		return
	}
	if uf.Version != ProtocolVersion {
		err = fmt.Errorf("unsupported version: %d, expected: %d", uf.Version, ProtocolVersion)
		return
	}
	err = uf.checkSourceAndTime()
	return
}

// stateful protocol interfaces
type StatefulDecoder struct {
	dec *json.Decoder
	Header
	Source
}

func NewStatefulDecoder(r io.Reader) (Decoder, error) {
	dec := &StatefulDecoder{}
	dec.dec = json.NewDecoder(r)
	if err := dec.dec.Decode(&dec); err != nil {
		return nil, err
	}
	if dec.Version != ProtocolVersion {
		return nil, fmt.Errorf("unsupported version: %d, expected: %d", dec.Version, ProtocolVersion)
	}
	return dec, nil
}

func (sd *StatefulDecoder) Slug() string {
	s := sd.Source
	return fmt.Sprintf("%s/%s/%s/%s", s.Hostname, s.Stream.ContentID, s.Stream.Format, s.Stream.Quality)
}

func (sd *StatefulDecoder) Decode() (uf *UpdateFull, err error) {
	uf = &UpdateFull{}
	uf.Version = sd.Version
	uf.CopyFromSource(&sd.Source)
	if err = sd.dec.Decode(uf); err != nil {
		return
	}
	if uf.Version != sd.Version {
		err = fmt.Errorf("unsupported version: %d, expected: %d", uf.Version, sd.Version)
		return
	}
	err = uf.checkSourceAndTime()
	return
}

//
// Encoder
//

type Encoder interface {
	Encode(*UpdateFull) error
	Slug() string
}

// stateless protocol interfaces
type StatelessEncoder struct {
	enc *json.Encoder
}

func NewStatelessEncoder(w io.Writer) Encoder {
	return &StatelessEncoder{json.NewEncoder(w)}
}

func (se *StatelessEncoder) Slug() string {
	return ""
}

func (se *StatelessEncoder) Encode(uf *UpdateFull) error {
	uf.Version = ProtocolVersion
	uf.StartTime = uf.StartTime.UTC()
	return se.enc.Encode(uf)
}

// stateful protocol interfaces
type StatefulEncoder struct {
	enc *json.Encoder
	Header
}

func NewStatefulEncoder(w io.Writer) (Encoder, error) {
	enc := &StatefulEncoder{}
	enc.Version = ProtocolVersion
	enc.enc = json.NewEncoder(w)
	return enc, enc.enc.Encode(&enc)
}

func (se *StatefulEncoder) Slug() string {
	return ""
}

func (se *StatefulEncoder) Encode(uf *UpdateFull) error {
	uf.Version = 0 // the init message took care of that
	uf.StartTime = uf.StartTime.UTC()
	return se.enc.Encode(uf)
}
