//
// sfive
//
// sfive - spreadspace streaming statistics suite is a generic
// statistic collection tool for streaming server infrastuctures.
// The system collects and stores meta data like number of views
// and throughput from a number of streaming servers and stores
// it in a global data store.
// The data acquisition is designed to be generic and extensible in
// order to support different streaming software.
// sfive also contains tools and applications to filter and visualize
// live and recorded data.
//
//
// Copyright (C) 2014-2017 Christian Pointner <equinox@spreadspace.org>
//                         Markus Grüneis <gimpf@gimpf.org>
//
// This file is part of sfive.
//
// sfive is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 3
// as published by the Free Software Foundation.
//
// sfive is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with sfive. If not, see <http://www.gnu.org/licenses/>.
//

package sfive

import (
	"bytes"
	"fmt"
	"reflect"
	"strings"
	"testing"
	"time"
)

var (
	initEncoded   = `"hostname": "localhost", "stream": {"quality": "low", "content": "av", "format": "webm"}, "tags": ["elevate", "2014"]`
	initStruct    = Source{Hostname: "localhost", Stream: Stream{Quality: "low", ContentID: "av", Format: "webm"}, Tags: []string{"elevate", "2014"}}
	updateEncoded = `"data": {"bytes-sent": 1, "client-count": 3, "bytes-received": 1}, "start-time": "2014-08-24T14:35:33.847282Z", "duration-ms": 5000`
	updateStruct  = Update{Data: UpdateData{BytesSent: 1, ClientCount: 3, BytesReceived: 1}, StartTime: time.Date(2014, time.August, 24, 14, 35, 33, 847282000, time.UTC), Duration: 5000}
)

func GetExpected() (expected *UpdateFull) {
	expected = &UpdateFull{}
	expected.Version = ProtocolVersion
	expected.CopyFromSource(&initStruct)
	expected.CopyFromUpdate(&updateStruct)
	return
}

func TestDecodeStateless(t *testing.T) {
	// invalid message
	dec := NewStatelessDecoder(strings.NewReader("this is still not json"))
	if _, err := dec.Decode(); err == nil {
		t.Fatalf("decoding invalid json should throw an error")
	}

	// check the slug
	if dec.Slug() != "" {
		t.Fatalf("the slug of a stateless decoder should be the empty string")
	}

	// wrong protocol version
	dec = NewStatelessDecoder(strings.NewReader(fmt.Sprintf(`{ "version": %d, %s, %s }`, ProtocolVersion+1, initEncoded, updateEncoded)))
	if _, err := dec.Decode(); err == nil {
		t.Fatalf("decoding message with wrong protocol version should throw an error")
	}

	// missing/empty source fields
	emptyHostname := `"hostname": "", "stream": {"quality": "low", "content": "av", "format": "webm"}`
	dec = NewStatelessDecoder(strings.NewReader(fmt.Sprintf(`{ "version": %d, %s, %s }`, ProtocolVersion, emptyHostname, updateEncoded)))
	if _, err := dec.Decode(); err == nil {
		t.Fatalf("decoding message with empty/missing hostname should throw an error")
	}
	emptyContentid := `"hostname": "localhost", "stream": {"quality": "low", "format": "webm"}`
	dec = NewStatelessDecoder(strings.NewReader(fmt.Sprintf(`{ "version": %d, %s, %s }`, ProtocolVersion, emptyContentid, updateEncoded)))
	if _, err := dec.Decode(); err == nil {
		t.Fatalf("decoding message with empty/missing content-id should throw an error")
	}
	emptyFormat := `"hostname": "localhost", "stream": {"quality": "low", "content": "av", "format": ""}`
	dec = NewStatelessDecoder(strings.NewReader(fmt.Sprintf(`{ "version": %d, %s, %s }`, ProtocolVersion, emptyFormat, updateEncoded)))
	if _, err := dec.Decode(); err == nil {
		t.Fatalf("decoding message with empty/missing format should throw an error")
	}
	emptyQuality := `"hostname": "localhost", "stream": {"content": "av", "format": "webm"}`
	dec = NewStatelessDecoder(strings.NewReader(fmt.Sprintf(`{ "version": %d, %s, %s }`, ProtocolVersion, emptyQuality, updateEncoded)))
	if _, err := dec.Decode(); err == nil {
		t.Fatalf("decoding message with empty/missing quality should throw an error")
	}

	// missing/invalid start-time and duration
	missingStartTime := `"data": {"bytes-sent": 1, "client-count": 3, "bytes-received": 1}, "duration-ms": 15000`
	dec = NewStatelessDecoder(strings.NewReader(fmt.Sprintf(`{ "version": %d, %s, %s }`, ProtocolVersion, initEncoded, missingStartTime)))
	if _, err := dec.Decode(); err == nil {
		t.Fatalf("decoding message with missing start-time should throw an error")
	}
	invalidStartTime := `"data": {"bytes-sent": 1, "client-count": 3, "bytes-received": 1}, "start-time": "this is not a date and time", "duration-ms": 15000`
	dec = NewStatelessDecoder(strings.NewReader(fmt.Sprintf(`{ "version": %d, %s, %s }`, ProtocolVersion, initEncoded, invalidStartTime)))
	if _, err := dec.Decode(); err == nil {
		t.Fatalf("decoding message with invalid start-time should throw an error")
	}
	missingDuration := `"data": {"bytes-sent": 1, "client-count": 3, "bytes-received": 1}, "start-time": "2014-08-24T14:35:33.847282Z"`
	dec = NewStatelessDecoder(strings.NewReader(fmt.Sprintf(`{ "version": %d, %s, %s }`, ProtocolVersion, initEncoded, missingDuration)))
	if _, err := dec.Decode(); err == nil {
		t.Fatalf("decoding message with missing duration should throw an error")
	}
	invalidDuration := `"data": {"bytes-sent": 1, "client-count": 3, "bytes-received": 1}, "start-time": "2014-08-24T14:35:33.847282Z", "duration-ms": -1`
	dec = NewStatelessDecoder(strings.NewReader(fmt.Sprintf(`{ "version": %d, %s, %s }`, ProtocolVersion, initEncoded, invalidDuration)))
	if _, err := dec.Decode(); err == nil {
		t.Fatalf("decoding message with invalid duration should throw an error")
	}

	// valid message
	dec = NewStatelessDecoder(strings.NewReader(fmt.Sprintf(`{ "version": %d, %s, %s }`, ProtocolVersion, initEncoded, updateEncoded)))
	decoded, err := dec.Decode()
	if err != nil {
		t.Fatalf("decoding message failed: %v", err)
	}

	// compare with expected result
	expected := GetExpected()
	if !reflect.DeepEqual(decoded, expected) {
		t.Fatalf("decoding failed:\n actual: %v\n expected: %v\n", decoded, expected)
	}
}

func TestDecodeStateful(t *testing.T) {
	// invalid init message
	if _, err := NewStatefulDecoder(strings.NewReader("this is not json")); err == nil {
		t.Fatalf("creating decoder with invalid json should throw an error")
	}

	// wrong protocol version
	wrongProtoVerMsg := fmt.Sprintf(`{ "version": %d, %s }`, ProtocolVersion+1, initEncoded)
	if _, err := NewStatefulDecoder(strings.NewReader(wrongProtoVerMsg)); err == nil {
		t.Fatalf("creating decoder with wrong protocol version should throw an error")
	}

	// valid init message
	statefulInitMsg := fmt.Sprintf(`{ "version": %d, %s }`, ProtocolVersion, initEncoded)
	if _, err := NewStatefulDecoder(strings.NewReader(statefulInitMsg)); err != nil {
		t.Fatalf("creating decoder failed: %v", err)
	}

	// invalid data-update message
	data := fmt.Sprintf(`{ "version": %d, %s }`, ProtocolVersion, initEncoded)
	data = data + "\nthis isn't valid json either"
	dec, err := NewStatefulDecoder(strings.NewReader(data))
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if _, err := dec.Decode(); err == nil {
		t.Fatalf("decoding message with invalid json should throw an error")
	}

	// wrong protocol version
	data = fmt.Sprintf(`{ "version": %d, %s }`, ProtocolVersion, initEncoded)
	data = data + fmt.Sprintf(`{ "version": %d, %s }`, ProtocolVersion+1, updateEncoded)
	dec, err = NewStatefulDecoder(strings.NewReader(data))
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if _, err := dec.Decode(); err == nil {
		t.Fatalf("decoding message with invalid json should throw an error")
	}

	// valid data-update message
	data = fmt.Sprintf(`{ "version": %d, %s }`, ProtocolVersion, initEncoded)
	data = data + fmt.Sprintf(`{ "version": %d, %s }`, ProtocolVersion, updateEncoded)
	dec, err = NewStatefulDecoder(strings.NewReader(data))
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	decoded, err := dec.Decode()
	if err != nil {
		t.Fatalf("decoding message failed: %v", err)
	}

	// compare with expected result
	expected := GetExpected()
	if !reflect.DeepEqual(decoded, expected) {
		t.Fatalf("decoding failed:\n actual: %v\n expected: %v\n", decoded, expected)
	}

	slug := dec.Slug()
	expectedSlug := strings.Join([]string{initStruct.Hostname, initStruct.Stream.ContentID,
		initStruct.Stream.Format, initStruct.Stream.Quality}, "/")
	if slug != expectedSlug {
		t.Fatalf("slug failed:\n actual: %v\n expected: %v\n", slug, expectedSlug)
	}

}

func TestEncodeStateless(t *testing.T) {
	uf := &UpdateFull{}
	uf.CopyFromSource(&initStruct)
	uf.CopyFromUpdate(&updateStruct)

	encoded := &bytes.Buffer{}
	enc := NewStatelessEncoder(encoded)
	if err := enc.Encode(uf); err != nil {
		t.Fatalf("encoding message failed: %v", err)
	}

	// check the slug
	if enc.Slug() != "" {
		t.Fatalf("the slug of a stateless encoder should be the empty string")
	}

	// try to decode encoded message
	decoded, err := NewStatelessDecoder(encoded).Decode()
	if err != nil {
		t.Fatalf("decoding message failed: %v", err)
	}

	expected := uf
	expected.Version = ProtocolVersion
	if !reflect.DeepEqual(decoded, expected) {
		t.Fatalf("encoding failed:\n actual: %v\n expected: %v\n", decoded, expected)
	}
}

func TestEncodeStateful(t *testing.T) {
	uf := &UpdateFull{}
	uf.CopyFromSource(&initStruct)
	uf.CopyFromUpdate(&updateStruct)

	encoded := &bytes.Buffer{}
	enc, err := NewStatefulEncoder(encoded)
	if err != nil {
		t.Fatalf("encoding init message failed: %v", err)
	}
	if err := enc.Encode(uf); err != nil {
		t.Fatalf("encoding data message failed: %v", err)
	}

	// check the slug
	if enc.Slug() != "" {
		t.Fatalf("the slug of a stateful encoder should be the empty string")
	}

	// try to decode encoded message
	dec, err := NewStatefulDecoder(encoded)
	if err != nil {
		t.Fatalf("decoding init message failed: %v", err)
	}
	decoded, err := dec.Decode()
	if err != nil {
		t.Fatalf("decoding data message failed: %v", err)
	}

	expected := uf
	expected.Version = ProtocolVersion
	if !reflect.DeepEqual(decoded, expected) {
		t.Fatalf("encoding failed:\n actual: %v\n expected: %v\n", decoded, expected)
	}
}
