//
// sfive
//
// sfive - spreadspace streaming statistics suite is a generic
// statistic collection tool for streaming server infrastuctures.
// The system collects and stores meta data like number of views
// and throughput from a number of streaming servers and stores
// it in a global data store.
// The data acquisition is designed to be generic and extensible in
// order to support different streaming software.
// sfive also contains tools and applications to filter and visualize
// live and recorded data.
//
//
// Copyright (C) 2014-2017 Christian Pointner <equinox@spreadspace.org>
//                         Markus Grüneis <gimpf@gimpf.org>
//
// This file is part of sfive.
//
// sfive is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 3
// as published by the Free Software Foundation.
//
// sfive is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with sfive. If not, see <http://www.gnu.org/licenses/>.
//

package sfive

import (
	"crypto/rand"
	"errors"
	"io/ioutil"
	"net"
	"os"

	"github.com/Yawning/cryptopan"
)

type AnonymizationAlgo interface {
	String() string
	Anonymize(addr string) (string, error)
}

func readOrGenerateKey(keyFile string, size int) (key []byte, err error) {
	if keyFile != "" {
		var kf *os.File
		if kf, err = os.Open(keyFile); err != nil {
			return
		}
		defer kf.Close()
		return ioutil.ReadAll(kf)
	}

	key = make([]byte, size)
	_, err = rand.Read(key)
	return
}

//
// CryptoPan based anonymization
//

type CryptoPanAnonymization struct {
	ctx     *cryptopan.Cryptopan
	keyFile string
}

func (cp *CryptoPanAnonymization) String() string {
	if cp.keyFile == "" {
		return "CryptoPan, with random key"
	}
	return "CryptoPan, with key from file: " + cp.keyFile
}

func (cp *CryptoPanAnonymization) Anonymize(addr string) (string, error) {
	ip := net.ParseIP(addr)
	if ip == nil {
		return "", errors.New("address '" + addr + "' is not valid")
	}
	return cp.ctx.Anonymize(ip).String(), nil
}

func NewCryptopanAnonymization(keyFile string) (AnonymizationAlgo, error) {
	key, err := readOrGenerateKey(keyFile, cryptopan.Size)
	if err != nil {
		return nil, err
	}

	cp := &CryptoPanAnonymization{keyFile: keyFile}
	if cp.ctx, err = cryptopan.New(key); err != nil {
		return nil, err
	}
	return cp, nil
}
