//
// sfive
//
// sfive - spreadspace streaming statistics suite is a generic
// statistic collection tool for streaming server infrastuctures.
// The system collects and stores meta data like number of views
// and throughput from a number of streaming servers and stores
// it in a global data store.
// The data acquisition is designed to be generic and extensible in
// order to support different streaming software.
// sfive also contains tools and applications to filter and visualize
// live and recorded data.
//
//
// Copyright (C) 2014-2017 Christian Pointner <equinox@spreadspace.org>
//                         Markus Grüneis <gimpf@gimpf.org>
//
// This file is part of sfive.
//
// sfive is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 3
// as published by the Free Software Foundation.
//
// sfive is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with sfive. If not, see <http://www.gnu.org/licenses/>.
//

package sfive

import (
	"bytes"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"net/http"
	"strings"
	"time"
)

const forwardEsLastUpdateIDJson = `{
  "query": {"match": { "SourceHubUuid": "%s" } },
  "aggregations": { "last-id" : { "max" : { "field": "SourceHubUpdateId" } } }
}`

func fwdEsGetLastUpdateID(baseUrl string, client *http.Client, hubUUID string) (lastID int64, err error) {
	url := baseUrl + "/dataupdate/_search?search_type=count"

	queryJson := fmt.Sprintf(forwardEsLastUpdateIDJson, hubUUID)
	s5dl.Printf("srv|fwd-es: query: %s", queryJson)

	var resp *http.Response
	resp, err = client.Post(url, "application/json", strings.NewReader(queryJson))
	if err != nil {
		s5l.Printf("srv|fwd-es: querying for lastupdate failed: %v", err)
		return
	}
	defer resp.Body.Close()

	if resp.StatusCode != http.StatusOK {
		s5l.Printf("srv|fwd-es: elasticsearch failed to fulfill query for lastupdate: %v", resp.StatusCode)
		return
	}

	var body []byte
	body, err = ioutil.ReadAll(resp.Body) // TODO: read from Body using json.Decoder
	if err != nil {
		s5l.Printf("srv|fwd-es: reading lastupdate response failed: %v", err)
		return
	}

	s5dl.Printf("srv|fwd-es: lastupdate response: %s", body)

	if len(body) == 0 {
		lastID = -1
	} else {
		var value interface{}
		err = json.Unmarshal(body, &value)
		if err != nil {
			s5l.Printf("srv|fwd-es: invalid lastupdate response: %v", err)
			return
		}

		idstrcntr := value.(map[string]interface{})["aggregations"].(map[string]interface{})["last-id"].(map[string]interface{})["value"]
		if idstrcntr == nil {
			s5l.Printf("srv|fwd-es: we are new here")
			lastID = -1
			return
		}
		tid := idstrcntr.(float64)
		lastID = int64(tid)
	}

	return
}

func (srv *Server) forwardEsRun(baseUrl string, client *http.Client) {
	url := baseUrl + "/_bulk"
	hubUUID := srv.store.GetHubUUID()
tryResync:
	for {
		lastID, err := fwdEsGetLastUpdateID(baseUrl, client, hubUUID)
		if err != nil {
			s5l.Printf("srv|fwd-es: fetching lastupdate failed: %v", err)
			time.Sleep(5 * time.Second)
			continue tryResync
		}
		s5l.Printf("srv|fwd-es: new lastupdate: %d", lastID)

	nextBatch:
		for {
			updates, err := srv.store.GetUpdatesAfter(lastID, 5000)
			if err != nil {
				s5l.Printf("srv|fwd-es: reading updates failed: %v", err)
				time.Sleep(500 * time.Millisecond)
				continue nextBatch
			}
			if len(updates) == 0 {
				time.Sleep(1 * time.Second)
				continue nextBatch
			}

			postData := bytes.Buffer{}

			for _, update := range updates {
				data, err := json.Marshal(update) // TODO: move to seperate function, use StatelessEncoder and write to io.PipeWriter
				if err != nil {
					s5l.Panicf("srv|fwd-es: encoding updates failed: %v", err)
				}
				postData.WriteString(`{"index":{"_type":"dataupdate"}}`)
				postData.WriteRune('\n')
				postData.Write(data)
				postData.WriteRune('\n')
			}

			// TODO: move to seperate function and use io.PipeReader as body
			resp, err := client.Post(url, "application/json", bytes.NewReader(postData.Bytes()))
			if err != nil {
				s5l.Printf("srv|fwd-es: posting updates failed: %v", err)
				time.Sleep(1 * time.Second)
				continue tryResync
			}

			if resp.StatusCode != http.StatusOK && resp.StatusCode != http.StatusCreated {
				body, _ := ioutil.ReadAll(resp.Body)
				s5l.Printf("srv|fwd-es: posting updates failed: %s\n%s", resp.Status, body)
				time.Sleep(1 * time.Second)
				continue tryResync
			}
			resp.Body.Close() // TODO: check result from elasticsearch

			lastID = findMaxID(updates)
			s5l.Printf("srv|fwd-es: successfully forwarded %d updates, new lastid: %d", len(updates), lastID)
		}
	}
}

func (srv *Server) RunForwardingEs(cfg ESForwardConfig) {
	s5l.Printf("srv|fwd-es: forwarding to '%s'", cfg.URL)
	defer s5l.Println("srv|fwd-es: forwarder stopped")

	srv.forwardEsRun(cfg.URL, http.DefaultClient)
}
