//
// sfive
//
// sfive - spreadspace streaming statistics suite is a generic
// statistic collection tool for streaming server infrastuctures.
// The system collects and stores meta data like number of views
// and throughput from a number of streaming servers and stores
// it in a global data store.
// The data acquisition is designed to be generic and extensible in
// order to support different streaming software.
// sfive also contains tools and applications to filter and visualize
// live and recorded data.
//
//
// Copyright (C) 2014-2017 Christian Pointner <equinox@spreadspace.org>
//                         Markus Grüneis <gimpf@gimpf.org>
//
// This file is part of sfive.
//
// sfive is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 3
// as published by the Free Software Foundation.
//
// sfive is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with sfive. If not, see <http://www.gnu.org/licenses/>.
//

package sfive

import (
	"context"
	"encoding/json"
	"errors"
	"io"
	"net/http"
	"time"
)

func findMaxID(updates []*UpdateFull) int64 {
	maxID := int64(-1)
	for _, value := range updates {
		if id := value.SourceHubUpdateID; id > maxID {
			maxID = id
		}
	}
	return maxID
}

func fwdGetLastUpdateID(baseUrl string, client *http.Client, hubUUID string) (lastID int64, err error) {
	lastID = -1

	var resp *http.Response
	resp, err = client.Get(baseUrl + "/api/v1/lastupdate/" + hubUUID)
	if err != nil {
		s5l.Printf("srv|fwd: querying for lastupdate failed: %v", err)
		return
	}
	defer resp.Body.Close()

	if resp.StatusCode != http.StatusOK {
		s5l.Printf("srv|fwd: remote hub failed to fulfill query for lastupdate: %v", resp.StatusCode)
		return
	}

	result := WebLastUpdateIDResponse{}
	if err = json.NewDecoder(resp.Body).Decode(&result); err != nil {
		s5l.Printf("srv|fwd: remote hub failed to fulfill query for lastupdate: %v", err)
		return
	}

	lastID = result.LastUpdateID
	return
}

func fwdWriteUpdates(updates []*UpdateFull, pw *io.PipeWriter) {
	defer pw.Close()

	enc, err := NewStatefulEncoder(pw)
	if err != nil {
		s5l.Printf("srv|fwd: encoding/sending init message failed: %v", err)
		return
	}
	for _, upd := range updates {
		if err := enc.Encode(upd); err != nil {
			s5l.Printf("srv|fwd: encoding/sending updates failed: %v", err)
			return
		}
	}
}

func fwdPostUpdates(client *http.Client, url string, pr *io.PipeReader) (int64, error) {
	resp, err := client.Post(url, "application/json", pr)
	if err != nil {
		return 0, err
	}
	defer resp.Body.Close()

	if resp.StatusCode != http.StatusOK {
		return 0, errors.New("")
	}
	result := WebUpdatesPostResponse{}
	if err = json.NewDecoder(resp.Body).Decode(&result); err != nil {
		return 0, err
	}
	return result.NumUpdates, nil
}

func (srv *Server) forwardRun(baseUrl string, client *http.Client) {
	url := baseUrl + "/api/v1/updates/_bulk"
	hubUUID := srv.store.GetHubUUID()
tryResync:
	for {
		lastID, err := fwdGetLastUpdateID(baseUrl, client, hubUUID)
		if err != nil {
			s5l.Printf("srv|fwd: fetching lastupdate failed: %v", err)
			time.Sleep(5 * time.Second)
			continue tryResync
		}
		s5l.Printf("srv|fwd: new lastupdate: %d", lastID)

	nextBatch:
		for {
			select {
			case ch := <-srv.forwarder.sfive:
				close(ch)
				return
			default:
			}

			updates, err := srv.store.GetUpdatesAfter(lastID, 5000)
			if err != nil {
				s5l.Printf("srv|fwd: reading updates failed: %v", err)
				time.Sleep(500 * time.Millisecond)
				continue nextBatch
			}
			if len(updates) == 0 {
				time.Sleep(1 * time.Second)
				continue nextBatch
			}

			pr, pw := io.Pipe()
			go fwdWriteUpdates(updates, pw)
			if num, err := fwdPostUpdates(client, url, pr); err != nil {
				s5l.Printf("srv|fwd: sending updates failed: %v", err)
				continue tryResync
			} else if num != int64(len(updates)) {
				s5l.Printf("srv|fwd: server acknowledged wrong number of updates: expected %d, got: %d", len(updates), num)
				continue tryResync
			}

			lastID = findMaxID(updates)
			s5l.Printf("srv|fwd: successfully forwarded %d updates, new lastid: %d", len(updates), lastID)
		}
	}
}

func (srv *Server) forwardStop(ctx context.Context) (err error) {
	if srv.forwarder.sfive == nil {
		return nil
	}
	s5l.Printf("srv|fwd: shutting down")
	stopCH := make(chan bool)
	srv.forwarder.sfive <- stopCH
	select {
	case <-stopCH:
	case <-ctx.Done():
	}
	return nil
}

func (srv *Server) ForwardSFive(cfg SFiveForwardConfig) {
	s5l.Printf("srv|fwd: forwarding to '%s'", cfg.URL)

	srv.wgShutdown.Add(1)
	go func() {
		defer srv.wgShutdown.Done()
		defer s5l.Println("srv|fwd: forwarder stopped")

		srv.forwarder.sfive = make(chan chan bool)
		srv.forwardRun(cfg.URL, http.DefaultClient)
	}()
}
