//
// sfive
//
// sfive - spreadspace streaming statistics suite is a generic
// statistic collection tool for streaming server infrastuctures.
// The system collects and stores meta data like number of views
// and throughput from a number of streaming servers and stores
// it in a global data store.
// The data acquisition is designed to be generic and extensible in
// order to support different streaming software.
// sfive also contains tools and applications to filter and visualize
// live and recorded data.
//
//
// Copyright (C) 2014-2017 Christian Pointner <equinox@spreadspace.org>
//                         Markus Grüneis <gimpf@gimpf.org>
//
// This file is part of sfive.
//
// sfive is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 3
// as published by the Free Software Foundation.
//
// sfive is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with sfive. If not, see <http://www.gnu.org/licenses/>.
//

package sfive

import (
	"context"
	"io"
	"net"
	"strings"
	"sync"
	"time"
)

func (srv *Server) pipeRead(dec Decoder, updateCH chan<- *UpdateFull, errorCH chan<- error) {
	for {
		update, err := dec.Decode()
		if err != nil {
			if err == io.EOF {
				close(updateCH)
				return
			}
			opErr, isOpErr := err.(*net.OpError)
			if isOpErr && opErr.Temporary() {
				continue
			}
			errorCH <- err
			return
		}
		updateCH <- update
	}
}

func (srv *Server) pipeHandle(conn net.Conn, quit <-chan bool) {
	defer conn.Close()

	conn.SetReadDeadline(time.Now().Add(10 * time.Second))
	dec, err := NewStatefulDecoder(conn)
	if err != nil {
		s5l.Printf("srv|pipe: read(init) failed: %v\n", err)
		return
	}
	conn.SetReadDeadline(time.Time{})

	slug := dec.Slug()
	s5l.Printf("srv|pipe: new connection: %s\n", slug)
	defer s5l.Printf("srv|pipe(%s): connection closed\n", slug)

	updateCH := make(chan *UpdateFull)
	errorCH := make(chan error)
	go srv.pipeRead(dec, updateCH, errorCH)

	for {
		select {
		case update := <-updateCH:
			if update == nil {
				return
			}
			if err = srv.Ingest(update); err != nil {
				s5l.Printf("srv|pipe(%s): storing data failed: %v\n", slug, err)
				// TODO: send NACK
				return
			}
			// TODO: send ACK
		case err := <-errorCH:
			s5l.Printf("srv|pipe(%s): read(data) failed: %v\n", slug, err)
			return
		case <-quit:
			return
		}
	}
}

func (srv *Server) pipeRun() {
	wgClients := &sync.WaitGroup{}
	quit := make(chan bool)
	for {
		conn, err := srv.interfaces.pipe.Accept()
		if err != nil {
			if strings.Contains(err.Error(), "use of closed network connection") { // TODO: is this really the best way to do this?
				break
			}
			s5l.Printf("srv|pipe: accept() failed: %v", err)
			// TODO: ignore ... or is this permanent?
			continue
		}
		wgClients.Add(1)
		go func() {
			defer wgClients.Done()
			srv.pipeHandle(conn, quit)
		}()
	}
	s5l.Println("srv|pipe: interface stopped listening")
	close(quit)
	wgClients.Wait()
}

func (srv *Server) pipeStop(ctx context.Context) (err error) {
	if srv.interfaces.pipe == nil {
		return nil
	}
	s5l.Printf("srv|pipe: shutting down")
	return srv.interfaces.pipe.Close()
}

func (srv *Server) ServePipe(cfg PipeInterfaceConfig) (err error) {
	if srv.interfaces.pipe, err = net.Listen("unix", cfg.ListenPath); err != nil {
		s5l.Printf("srv|pipe: listen() failed: %v", err)
		return
	}

	s5l.Printf("srv|pipe: listening on '%s'", cfg.ListenPath)

	srv.wgShutdown.Add(1)
	go func() {
		defer srv.wgShutdown.Done()

		srv.pipeRun()
	}()
	return
}
