//
// sfive
//
// sfive - spreadspace streaming statistics suite is a generic
// statistic collection tool for streaming server infrastuctures.
// The system collects and stores meta data like number of views
// and throughput from a number of streaming servers and stores
// it in a global data store.
// The data acquisition is designed to be generic and extensible in
// order to support different streaming software.
// sfive also contains tools and applications to filter and visualize
// live and recorded data.
//
//
// Copyright (C) 2014-2017 Christian Pointner <equinox@spreadspace.org>
//                         Markus Grüneis <gimpf@gimpf.org>
//
// This file is part of sfive.
//
// sfive is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 3
// as published by the Free Software Foundation.
//
// sfive is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with sfive. If not, see <http://www.gnu.org/licenses/>.
//

package sfive

import (
	"context"
	"encoding/json"
	"fmt"
	"io"
	"net"
	"net/http"
	"path"
	"strconv"
	"strings"
	"time"
)

func webNotFound(srv *Server, w http.ResponseWriter, r *http.Request) {
	// TODO: show index on '^/$'
	sendWebResponse(w, http.StatusNotFound, WebErrorResponse{"not found"})
}

//
// /healthz
//

func webHealthz(srv *Server, w http.ResponseWriter, r *http.Request) {
	if r.Method != "GET" {
		sendInvalidMethod(w, r.Method)
		return
	}

	resp := WebHealthzResponse{}
	resp.HubUUID = srv.store.GetHubUUID()
	resp.Status = "OK" // TODO: do a more sophisticated check
	sendWebResponse(w, http.StatusOK, resp)
}

//
// /hubs
//

func webHubs(srv *Server, w http.ResponseWriter, r *http.Request) {
	if r.Method != "GET" {
		sendInvalidMethod(w, r.Method)
		return
	}

	var err error
	resp := WebHubsResponse{}
	if resp.Hubs, err = srv.store.GetHubs(); err != nil {
		sendWebResponse(w, http.StatusInternalServerError, WebErrorResponse{err.Error()})
		return
	}
	sendWebResponse(w, http.StatusOK, resp)
}

//
// /sources
//

func webSources(srv *Server, w http.ResponseWriter, r *http.Request) {
	if r.Method != "GET" {
		sendInvalidMethod(w, r.Method)
		return
	}

	var err error
	resp := WebSourcesResponse{}
	if resp.Sources, err = srv.store.GetSources(); err != nil {
		sendWebResponse(w, http.StatusInternalServerError, WebErrorResponse{err.Error()})
		return
	}
	sendWebResponse(w, http.StatusOK, resp)
}

//
// /clients
//

func webClients(srv *Server, w http.ResponseWriter, r *http.Request) {
	if r.Method != "GET" {
		sendInvalidMethod(w, r.Method)
		return
	}

	var err error
	resp := WebClientsResponse{}
	if resp.Clients, err = srv.store.GetClients(); err != nil {
		sendWebResponse(w, http.StatusInternalServerError, WebErrorResponse{err.Error()})
		return
	}
	sendWebResponse(w, http.StatusOK, resp)
}

//
// /updates/(:ID|_bulk)
//

func webUpdatesWithParam(srv *Server, w http.ResponseWriter, r *http.Request) {
	bulk := r.URL.Path == "/updates/_bulk"

	switch r.Method {
	case "GET":
		if bulk {
			webUpdatesGetBulk(srv, w, r)
			return
		}

		if matched, err := path.Match("/updates/?*", r.URL.Path); err != nil || !matched {
			sendWebResponse(w, http.StatusBadRequest, WebErrorResponse{"invalid update id: " + strings.TrimPrefix("/updates/", r.URL.Path)})
			return
		}
		id, err := strconv.ParseInt(path.Base(r.URL.Path), 10, 64)
		if err != nil {
			sendWebResponse(w, http.StatusBadRequest, WebErrorResponse{"invalid update id: " + err.Error()})
			return
		}
		webUpdateGet(srv, id, w, r)
	case "POST":
		if bulk {
			webUpdatesPostBulk(srv, w, r)
			return
		}

		sendWebResponse(w, http.StatusBadRequest, WebErrorResponse{"invalid parameter: " + strings.TrimPrefix("/updates/", r.URL.Path)})
	default:
		sendInvalidMethod(w, r.Method)
	}
}

//
// /updates
//

func webUpdates(srv *Server, w http.ResponseWriter, r *http.Request) {
	switch r.Method {
	case "GET":
		webUpdatesGet(srv, w, r)
	case "POST":
		webUpdatePost(srv, w, r)
	default:
		sendInvalidMethod(w, r.Method)
	}
}

// get one update

func webUpdateGet(srv *Server, id int64, w http.ResponseWriter, r *http.Request) {
	update, err := srv.store.GetUpdate(id)
	if err != nil {
		status := http.StatusInternalServerError
		if err == ErrNotFound {
			status = http.StatusNotFound
		}
		sendWebResponse(w, status, WebErrorResponse{err.Error()})
		return
	}

	sendWebResponse(w, http.StatusOK, update)
}

// get list of updates (as json array)

func webUpdatesGet(srv *Server, w http.ResponseWriter, r *http.Request) {
	after, limit, err := parseAfterAndLimit(r)
	if err != nil {
		sendWebResponse(w, http.StatusBadRequest, WebErrorResponse{err.Error()})
	}

	resp := WebUpdatesGetResponse{}
	if resp.Updates, err = srv.store.GetUpdatesAfter(after, limit); err != nil {
		sendWebResponse(w, http.StatusInternalServerError, WebErrorResponse{err.Error()})
		return
	}
	sendWebResponse(w, http.StatusOK, resp)
}

// get list of updates (bulk format)

func webUpdatesGetBulk(srv *Server, w http.ResponseWriter, r *http.Request) {
	after, limit, err := parseAfterAndLimit(r)
	if err != nil {
		sendWebResponse(w, http.StatusBadRequest, WebErrorResponse{err.Error()})
	}

	updates, err := srv.store.GetUpdatesAfter(after, limit)
	if err != nil {
		sendWebResponse(w, http.StatusInternalServerError, WebErrorResponse{err.Error()})
		return
	}

	w.Header().Set("Content-Type", "application/json") // this is actually multiple json documents...
	w.WriteHeader(http.StatusOK)

	enc, err := NewStatefulEncoder(w)
	if err != nil {
		s5l.Printf("srv|web: sending response(init) failed: %v", err)
		return
	}
	for _, update := range updates {
		if err := enc.Encode(update); err != nil {
			s5l.Printf("srv|web: sending response(data) failed: %v", err)
			return
		}
	}
}

// post one update

func webUpdatePost(srv *Server, w http.ResponseWriter, r *http.Request) {
	dec := NewStatelessDecoder(r.Body)
	update, err := dec.Decode()
	if err != nil {
		sendWebResponse(w, http.StatusBadRequest, WebErrorResponse{"error while decoding update: " + err.Error()})
		return
	}

	if err = srv.Ingest(update); err != nil {
		sendWebResponse(w, http.StatusInternalServerError, WebErrorResponse{err.Error()})
		return
	}

	sendWebResponse(w, http.StatusOK, WebUpdatesPostResponse{1})
}

// post multiple updates in bulk mode

func webUpdatesPostBulk(srv *Server, w http.ResponseWriter, r *http.Request) {
	dec, err := NewStatefulDecoder(r.Body)
	if err != nil && err != io.EOF {
		sendWebResponse(w, http.StatusBadRequest, WebErrorResponse{"failed to read/decode init message: " + err.Error()})
		return
	}

	updates := []*UpdateFull{}
	for {
		update, err := dec.Decode()
		if err != nil {
			if err != io.EOF {
				sendWebResponse(w, http.StatusBadRequest, WebErrorResponse{"failed to read/decode data message: " + err.Error()})
				return
			}
			break
		}
		updates = append(updates, update)
	}
	numUpdates := int64(len(updates))
	if numUpdates < 1 {
		sendWebResponse(w, http.StatusBadRequest, WebErrorResponse{"got no data messages"})
		return
	}

	if err = srv.IngestMany(updates); err != nil {
		sendWebResponse(w, http.StatusInternalServerError, WebErrorResponse{err.Error()})
		return
	}

	sendWebResponse(w, http.StatusOK, WebUpdatesPostResponse{numUpdates})
}

//
// /lastupdate/:UUID
//

func webLastUpdateIDForUUID(srv *Server, w http.ResponseWriter, r *http.Request) {
	if r.Method != "GET" {
		sendInvalidMethod(w, r.Method)
		return
	}

	if matched, err := path.Match("/lastupdate/?*", r.URL.Path); err != nil || !matched {
		sendWebResponse(w, http.StatusBadRequest, WebErrorResponse{"invalid uuid"})
		return
	}

	var err error
	resp := WebLastUpdateIDResponse{}
	resp.HubUUID = path.Base(r.URL.Path)
	if resp.LastUpdateID, err = srv.store.GetLastUpdateIDForUUID(resp.HubUUID); err != nil {
		sendWebResponse(w, http.StatusInternalServerError, WebErrorResponse{err.Error()})
		return
	}
	sendWebResponse(w, http.StatusOK, resp)
}

//
// /lastupdate
//

func webLastUpdateID(srv *Server, w http.ResponseWriter, r *http.Request) {
	if r.Method != "GET" {
		sendInvalidMethod(w, r.Method)
		return
	}

	var err error
	resp := WebLastUpdateIDResponse{}
	resp.HubUUID = srv.store.GetHubUUID()
	if resp.LastUpdateID, err = srv.store.GetLastUpdateID(); err != nil {
		sendWebResponse(w, http.StatusInternalServerError, WebErrorResponse{err.Error()})
		return
	}
	sendWebResponse(w, http.StatusOK, resp)
}

//
// common functions
//

func parseAfterAndLimit(r *http.Request) (after, limit int64, err error) {
	q := r.URL.Query()

	afterStr := q.Get("after")
	if afterStr != "" {
		if after, err = strconv.ParseInt(afterStr, 10, 64); err != nil {
			err = fmt.Errorf("invalid after value: %v", err.Error())
			return
		}
	}

	limit = -1
	limitStr := q.Get("limit")
	if limitStr != "" {
		if limit, err = strconv.ParseInt(limitStr, 10, 64); err != nil {
			err = fmt.Errorf("invalid limit value: %v", err.Error())
			return
		}
	}

	return
}

func sendInvalidMethod(w http.ResponseWriter, method string) {
	sendWebResponse(w, http.StatusMethodNotAllowed, WebErrorResponse{"invalid request method: " + method})
}

func sendWebResponse(w http.ResponseWriter, status int, respdata interface{}) {
	w.Header().Set("Content-Type", "application/json")
	w.WriteHeader(status)
	if err := json.NewEncoder(w).Encode(respdata); err != nil {
		s5l.Printf("srv|web: sending response failed: %v", err)
	}
}

type webHandler struct {
	srv *Server
	H   func(*Server, http.ResponseWriter, *http.Request)
}

func (wh webHandler) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	wh.H(wh.srv, w, r)
}

// This is from golang http package - why is this not exported?
type tcpKeepAliveListener struct {
	*net.TCPListener
}

func (ln tcpKeepAliveListener) Accept() (c net.Conn, err error) {
	tc, err := ln.AcceptTCP()
	if err != nil {
		return
	}
	tc.SetKeepAlive(true)
	tc.SetKeepAlivePeriod(30 * time.Second)
	return tc, nil
}

func (srv *Server) webRun(listener *net.TCPListener) {
	srv.interfaces.web.Serve(tcpKeepAliveListener{listener})
}

func (srv *Server) webStop(ctx context.Context) (err error) {
	if srv.interfaces.web == nil {
		return nil
	}
	s5l.Printf("srv|web: shutting down")
	return srv.interfaces.web.Shutdown(ctx)
}

func (srv *Server) ServeWeb(cfg WebInterfaceConfig) error {
	ln, err := net.Listen("tcp", cfg.ListenAddr)
	if err != nil {
		s5l.Printf("srv|web: listen() failed: %v", err)
		return err
	}

	s5l.Printf("srv|web: listening on '%s'", cfg.ListenAddr)

	mux := http.NewServeMux()

	mux.Handle("/api/v1/hubs", http.StripPrefix("/api/v1", webHandler{srv, webHubs}))
	mux.Handle("/api/v1/sources", http.StripPrefix("/api/v1", webHandler{srv, webSources}))
	mux.Handle("/api/v1/clients", http.StripPrefix("/api/v1", webHandler{srv, webClients}))
	mux.Handle("/api/v1/updates/", http.StripPrefix("/api/v1", webHandler{srv, webUpdatesWithParam}))
	mux.Handle("/api/v1/updates", http.StripPrefix("/api/v1", webHandler{srv, webUpdates}))
	mux.Handle("/api/v1/lastupdate/", http.StripPrefix("/api/v1", webHandler{srv, webLastUpdateIDForUUID}))
	mux.Handle("/api/v1/lastupdate", http.StripPrefix("/api/v1", webHandler{srv, webLastUpdateID}))

	mux.Handle("/healthz", webHandler{srv, webHealthz})
	//	mux.Handle("/static/", http.StripPrefix("/static/", http.FileServer(http.Dir( ..staticDir.. ))))
	mux.Handle("/", webHandler{srv, webNotFound})

	srv.interfaces.web = &http.Server{Handler: mux, ReadTimeout: 60 * time.Second, WriteTimeout: 60 * time.Second}

	srv.wgShutdown.Add(1)
	go func() {
		defer srv.wgShutdown.Done()
		defer s5l.Println("srv|web: interface stopped listening")

		srv.webRun(ln.(*net.TCPListener))
	}()
	return nil
}
