//
// sfive
//
// sfive - spreadspace streaming statistics suite is a generic
// statistic collection tool for streaming server infrastuctures.
// The system collects and stores meta data like number of views
// and throughput from a number of streaming servers and stores
// it in a global data store.
// The data acquisition is designed to be generic and extensible in
// order to support different streaming software.
// sfive also contains tools and applications to filter and visualize
// live and recorded data.
//
//
// Copyright (C) 2014-2017 Christian Pointner <equinox@spreadspace.org>
//                         Markus Grüneis <gimpf@gimpf.org>
//
// This file is part of sfive.
//
// sfive is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 3
// as published by the Free Software Foundation.
//
// sfive is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with sfive. If not, see <http://www.gnu.org/licenses/>.
//

package sfive

import (
	"fmt"
	"io"
	"os"
	"os/user"
	"reflect"
	"testing"
	"time"

	"github.com/coreos/bbolt"
)

var (
	testBoltPath      = "/run/s5hub_testing_db.bolt"
	testBoltPath2     = "/run/s5hub_testing_db2.bolt"
	testBoltPathFwd   = "/run/s5hub_testing_db_fwd.bolt"
	testBoltPathFinal = "/run/s5hub_testing_db_final.bolt"

	streamIDData = Stream{ContentID: "talkingheads", Format: "7bitascii", Quality: "high"}
	sourceData   = Source{Hostname: "streamer", Tags: []string{"tag1", "master"}, Stream: streamIDData}
	updateData   = Update{Data: UpdateData{ClientCount: 3, BytesReceived: 42, BytesSent: 136}, Duration: 5000}
	clientsData  = []Client{
		Client{"127.0.0.1", 51236, "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:53.0) Gecko/20100101 Firefox/53.0", 6400, GeoInfo{}},
		Client{"10.12.0.1", 0, "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/57.0.2987.98 Chrome/57.0.2987.98 Safari/537.36", 6400, GeoInfo{}},
		Client{"127.0.0.1", 51202, "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:53.0) Gecko/20100101 Firefox/53.0", 6400, GeoInfo{}},
		Client{"192.168.0.1", 49487, "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/57.0.2987.98 Chrome/57.0.2987.98 Safari/537.36", 6400, GeoInfo{}},
		Client{"172.16.0.2", 0, "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/57.0.2987.98 Chrome/57.0.2987.98 Safari/537.36", 6400, GeoInfo{}}}
)

func generateTestData(n int) ([]*UpdateFull, int, int) {
	hostnames := []string{"streamer1", "streamer2"}
	contents := []string{"av", "audio"}
	formats := []string{"webm", "flash", "hls"}
	qualities := []string{"high", "medium", "low"}
	tags := []string{"2017", "elevate", "discourse"}
	duration := int64(15000)

	starttime := time.Now()
	numSrcs := len(hostnames) * len(contents) * len(formats) * len(qualities)
	if n < 0 {
		n = numSrcs
	}
	var data []*UpdateFull
	for i := 0; i < n; i += numSrcs {
		for _, hostname := range hostnames {
			for _, content := range contents {
				for _, format := range formats {
					for _, quality := range qualities {
						d := &UpdateFull{}
						d.Version = ProtocolVersion

						d.Source.Hostname = hostname
						d.Source.Stream.ContentID = content
						d.Source.Stream.Format = format
						d.Source.Stream.Quality = quality
						d.Source.Tags = tags

						d.Update.StartTime = starttime
						d.Update.Duration = duration
						d.Update.Data.ClientCount = uint(len(clientsData))
						d.Update.Data.BytesSent = 6400 * uint(len(clientsData))

						d.Update.Data.Clients = clientsData
						data = append(data, d)
					}
				}
			}
		}
		starttime = starttime.Add(time.Duration(duration) * time.Millisecond)
	}
	return data[:n], numSrcs, len(clientsData)
}

func TestMain(m *testing.M) {
	u, err := user.Current()
	if err != nil {
		os.Exit(-1)
	}
	testBoltPath = fmt.Sprintf("/run/user/%s/s5hub_testing_db.bolt", u.Uid)
	testBoltPath2 = fmt.Sprintf("/run/user/%s/s5hub_testing_db2.bolt", u.Uid)
	testBoltPathFwd = fmt.Sprintf("/run/user/%s/s5hub_testing_db_fwd.bolt", u.Uid)
	testBoltPathFinal = fmt.Sprintf("/run/user/%s/s5hub_testing_db_final.bolt", u.Uid)
	os.Exit(m.Run())
}

//
// Testing
//

func TestIDList(t *testing.T) {
	var ids idList

	b := ids.toBytes()
	if len(b) != 0 {
		t.Fatalf("an empty id-list should yield an empty byte array but returned: %v", b)
	}

	ids = append(ids, 12, 2198319823423, 4212411244)
	b = ids.toBytes()
	if len(b) != 3*8 {
		t.Fatalf("an id-list with 3 entries should return a 24 bytes but returned: %v", b)
	}
	ids2, err := idListFromBytes(b)
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if !reflect.DeepEqual([]int64(ids2), []int64(ids)) {
		t.Fatalf("failed to read id-list, got: %v, expected: %v", ids2, ids)
	}

	b = appendToBinaryIDList(b, 23)
	if ids2, err = idListFromBytes(b); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	ids = append(ids, 23)
	if !reflect.DeepEqual([]int64(ids2), []int64(ids)) {
		t.Fatalf("failed to read id-list, got: %v, expected: %v", ids2, ids)
	}

	// some more toBytes tests...
	if ids2, err = idListFromBytes([]byte{}); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if ids2, err = idListFromBytes([]byte{124}); err == nil {
		t.Fatalf("read id-list from bytes with wrong length should return an error")
	}

	if ids2, err = idListFromBytes([]byte{0, 0, 0, 0, 0, 0, 0, 1}); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	expected := []int64{1}
	if !reflect.DeepEqual([]int64(ids2), expected) {
		t.Fatalf("failed to read id-list, got: %v, expected: %v", ids2, expected)
	}
}

func TestOpen(t *testing.T) {
	// non-existing directory
	if _, err := NewStore(StoreConfig{"/nonexistend/db.bolt", false}); err == nil {
		t.Fatalf("opening store in nonexisting directory should throw an error")
	}

	// store path is a directory
	os.Remove(testBoltPath)
	if err := os.MkdirAll(testBoltPath, 0700); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if _, err := NewStore(StoreConfig{testBoltPath, false}); err == nil {
		t.Fatalf("opening store using a directory should throw an error")
	}

	// exisiting but non-database file
	os.Remove(testBoltPath)
	if f, err := os.Create(testBoltPath); err != nil {
		t.Fatalf("unexpected error: %v", err)
	} else {
		io.WriteString(f, "this is not a bolt db.")
		f.Close()
	}
	if _, err := NewStore(StoreConfig{testBoltPath, false}); err == nil {
		t.Fatalf("opening store using a invalid database should throw an error")
	}

	// bolt db without HubInfo Bucket
	os.Remove(testBoltPath)
	db, err := bolt.Open(testBoltPath, 0600, &bolt.Options{Timeout: 100 * time.Millisecond})
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	} else {
		if err = db.Close(); err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
	}
	if _, err := NewStore(StoreConfig{testBoltPath, false}); err == nil {
		t.Fatalf("opening store without HubInfo Bucket should throw an error")
	}

	// bolt db no version key
	if db, err := bolt.Open(testBoltPath, 0600, &bolt.Options{Timeout: 100 * time.Millisecond}); err != nil {
		t.Fatalf("unexpected error: %v", err)
	} else {
		err = db.Update(func(tx *bolt.Tx) error {
			if _, err := tx.CreateBucket([]byte(hubInfoBn)); err != nil {
				return err
			}
			return nil
		})
		if err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
		if err = db.Close(); err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
	}
	if _, err := NewStore(StoreConfig{testBoltPath, false}); err == nil {
		t.Fatalf("opening store without a database version should throw an error")
	}

	// bolt db wrong version
	if db, err := bolt.Open(testBoltPath, 0600, &bolt.Options{Timeout: 100 * time.Millisecond}); err != nil {
		t.Fatalf("unexpected error: %v", err)
	} else {
		err = db.Update(func(tx *bolt.Tx) error {
			b := tx.Bucket([]byte(hubInfoBn))
			if err := b.Put([]byte(storeVersionKey), itob(0)); err != nil {
				return err
			}
			return nil
		})
		if err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
		if err = db.Close(); err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
	}
	if _, err := NewStore(StoreConfig{testBoltPath, false}); err == nil {
		t.Fatalf("opening store with wrong database version should throw an error")
	}

	// bolt db no UUID key
	if db, err := bolt.Open(testBoltPath, 0600, &bolt.Options{Timeout: 100 * time.Millisecond}); err != nil {
		t.Fatalf("unexpected error: %v", err)
	} else {
		err = db.Update(func(tx *bolt.Tx) error {
			b := tx.Bucket([]byte(hubInfoBn))
			if err := b.Put([]byte(storeVersionKey), itob(StoreVersion)); err != nil {
				return err
			}
			return nil
		})
		if err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
		if err = db.Close(); err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
	}
	if _, err := NewStore(StoreConfig{testBoltPath, false}); err == nil {
		t.Fatalf("opening store without a database UUID should throw an error")
	}

	// bolt db empty UUID
	if db, err := bolt.Open(testBoltPath, 0600, &bolt.Options{Timeout: 100 * time.Millisecond}); err != nil {
		t.Fatalf("unexpected error: %v", err)
	} else {
		err = db.Update(func(tx *bolt.Tx) error {
			b := tx.Bucket([]byte(hubInfoBn))
			if err := b.Put([]byte(hubUUIDKey), []byte("")); err != nil {
				return err
			}
			return nil
		})
		if err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
		if err = db.Close(); err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
	}
	if _, err := NewStore(StoreConfig{testBoltPath, false}); err == nil {
		t.Fatalf("opening store with empty UUID should throw an error")
	}

	// bolt db with missing buckets
	if db, err := bolt.Open(testBoltPath, 0600, &bolt.Options{Timeout: 100 * time.Millisecond}); err != nil {
		t.Fatalf("unexpected error: %v", err)
	} else {
		err = db.Update(func(tx *bolt.Tx) error {
			b := tx.Bucket([]byte(hubInfoBn))
			if err := b.Put([]byte(hubUUIDKey), []byte("6d1a2192-7cb6-404d-80fb-add9b40a8f33")); err != nil {
				return err
			}
			return nil
		})
		if err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
		if err = db.Close(); err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
	}
	if _, err := NewStore(StoreConfig{testBoltPath, false}); err == nil {
		t.Fatalf("opening store with (some) buckets missing should throw an error")
	}

	// bolt db with missing index buckets
	if db, err := bolt.Open(testBoltPath, 0600, &bolt.Options{Timeout: 100 * time.Millisecond}); err != nil {
		t.Fatalf("unexpected error: %v", err)
	} else {
		err = db.Update(func(tx *bolt.Tx) error {
			for _, bn := range storeBuckets {
				if _, err := tx.CreateBucket([]byte(bn)); err != nil {
					return err
				}
			}
			return nil
		})
		if err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
		if err = db.Close(); err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
	}
	if _, err := NewStore(StoreConfig{testBoltPath, false}); err == nil {
		t.Fatalf("opening store with index buckets missing should throw an error")
	}

	// create new bolt-db and reopen it
	os.Remove(testBoltPath)
	store, err := NewStore(StoreConfig{testBoltPath, false})
	if err != nil {
		t.Fatalf("creating new store failed: %v", err)
	}
	createdUUID := store.hubUUID
	store.Close()

	store, err = NewStore(StoreConfig{testBoltPath, false})
	if err != nil {
		t.Fatalf("re-opening existing store failed: %v", err)
	}
	if createdUUID != store.hubUUID {
		t.Fatalf("UUID of opened store differs from the one previously generated: '%s' != '%s'", createdUUID, store.hubUUID)
	}

	if _, err := NewStore(StoreConfig{testBoltPath, false}); err == nil {
		t.Fatalf("opening already opened database should throw an error")
	}
	store.Close()
}

func TestAppendAndFetch(t *testing.T) {
	os.Remove(testBoltPath)
	store, err := NewStore(StoreConfig{testBoltPath, false})
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	defer store.Close()

	if _, err := store.GetUpdate(17); err != ErrNotFound {
		t.Fatalf("fetching not exisiting update should return not-found, error=%v", err)
	}

	upd := updateData
	upd.StartTime = time.Date(2014, time.August, 24, 14, 35, 33, 847000000, time.UTC)
	upd.Data.Clients = clientsData
	in := &UpdateFull{Header{0, "", -1, "", -1}, sourceData, upd}

	if err = store.Append(in); err != nil {
		t.Fatalf("failed to append update: %v", err)
	}

	out, err := store.GetUpdate(1)
	if err != nil {
		t.Fatalf("failed to fetch update: %v", err)
	}
	out.StartTime = out.StartTime.UTC() // this would normally be handled by the protocol encoder

	expected := in
	expected.SourceHubUUID = store.GetHubUUID()
	expected.SourceHubUpdateID = 1
	expected.ForwardHubUUID = ""
	expected.ForwardHubUpdateID = 0

	if !reflect.DeepEqual(expected, out) {
		t.Fatalf("failed to fetch update\nactual: %v\nexpected: %v\n", out, expected)
	}

	// append many
	var ins []*UpdateFull
	upd.StartTime = upd.StartTime.Add(time.Duration(upd.Duration) * time.Millisecond)
	ins = append(ins, &UpdateFull{Header{0, "", -1, "", -1}, sourceData, upd})
	upd.StartTime = upd.StartTime.Add(time.Duration(upd.Duration) * time.Millisecond)
	upd.Data.Clients = nil
	ins = append(ins, &UpdateFull{Header{0, "", -1, "", -1}, sourceData, upd})
	if err = store.AppendMany(ins); err != nil {
		t.Fatalf("failed to append update: %v", err)
	}

	for i := int64(0); i < 2; i = i + 1 {
		out, err = store.GetUpdate(i + 2)
		if err != nil {
			t.Fatalf("failed to fetch update: %v", err)

		}
		out.StartTime = out.StartTime.UTC() // this would normally be handled by the protocol encoder
		expected = ins[i]
		expected.SourceHubUUID = store.GetHubUUID()
		expected.SourceHubUpdateID = i + 2
		expected.ForwardHubUUID = ""
		expected.ForwardHubUpdateID = 0

		if !reflect.DeepEqual(expected, out) {
			t.Fatalf("failed to fetch update\nactual: %v\nexpected: %v\n", out, expected)
		}
	}
}

func TestTimestampIndex(t *testing.T) {
	os.Remove(testBoltPath)
	store, err := NewStore(StoreConfig{testBoltPath, false})
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	defer store.Close()

	upd := updateData
	upd.StartTime = time.Date(2014, time.August, 24, 14, 35, 33, 847000000, time.UTC)
	upd.Data.Clients = clientsData
	in := &UpdateFull{Header{0, "", -1, "", -1}, sourceData, upd}

	if err = store.Append(in); err != nil {
		t.Fatalf("failed to append update: %v", err)
	}
	if err = store.Append(in); err != nil {
		t.Fatalf("failed to append update: %v", err)
	}
	in.Update.StartTime = in.Update.StartTime.Add(1 * time.Hour)
	if err = store.Append(in); err != nil {
		t.Fatalf("failed to append update: %v", err)
	}
	in.Update.StartTime = in.Update.StartTime.Add(24 * time.Hour)
	if err = store.Append(in); err != nil {
		t.Fatalf("failed to append update: %v", err)
	}

	err = store.db.View(func(tx *bolt.Tx) error {
		tsb := tx.Bucket([]byte(timestampsIdxBn))

		db24 := tsb.Bucket([]byte("2014-08-24"))
		if db24 == nil {
			t.Fatalf("the append should have created an bucket for 24.8.2014")
		}
		db25 := tsb.Bucket([]byte("2014-08-25"))
		if db25 == nil {
			t.Fatalf("the append should have created an bucket for 25.8.2014")
		}

		// TODO: check if db24 and db25 contain the correct entries

		return nil
	})
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
}

func TestReadOnly(t *testing.T) {
	// create read-only db from not-existing file must fail
	os.Remove(testBoltPath)
	if _, err := NewStore(StoreConfig{testBoltPath, true}); err == nil {
		t.Fatalf("creating a read-only database should throw an error")
	}

	// prepare a store with one data-update
	store, err := NewStore(StoreConfig{testBoltPath, false})
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}

	upd := updateData
	upd.StartTime = time.Date(2014, time.August, 24, 14, 35, 33, 847000000, time.UTC)
	upd.Data.Clients = clientsData
	in := &UpdateFull{Header{0, "", -1, "", -1}, sourceData, upd}
	if err = store.Append(in); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	store.Close()

	// read-only db from existing file must succeed
	store, err = NewStore(StoreConfig{testBoltPath, true})
	if err != nil {
		t.Fatalf("opening existing store in read-only mode failed: %v", err)
	}
	defer store.Close()

	// fetching the date-update from read-only store must succeed
	if _, err := store.GetUpdate(1); err != nil {
		t.Fatalf("failed to fetch update: %v", err)
	}

	// appending to read-only store must fail
	if err = store.Append(in); err == nil {
		t.Fatalf("appending to read-only store should throw an error")
	}
}

func TestGetUpdatesAfter(t *testing.T) {
	// prepare a store with 3 data-updates
	os.Remove(testBoltPath)
	store, err := NewStore(StoreConfig{testBoltPath, false})
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	defer store.Close()

	upd := updateData
	upd.StartTime = time.Date(2014, time.August, 24, 14, 35, 33, 847000000, time.UTC)
	upd.Data.Clients = clientsData

	expected := []*UpdateFull{}
	for i := int64(0); i < 3; i = i + 1 {
		in := &UpdateFull{Header{0, "", -1, "", -1}, sourceData, upd}
		if err = store.Append(in); err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
		e := in
		e.SourceHubUUID = store.hubUUID
		e.SourceHubUpdateID = i + 1
		e.ForwardHubUUID = ""
		e.ForwardHubUpdateID = 0
		expected = append(expected, e)
		upd.StartTime = upd.StartTime.Add(time.Duration(upd.Duration) * time.Millisecond)
	}

	// check if there are 3 updates in store
	lastID, err := store.GetLastUpdateID()
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if lastID != 3 {
		t.Fatalf("failed to get last update ID: got %d updates, expected 3", lastID)
	}

	lastIDForUUID, err := store.GetLastUpdateIDForUUID(store.hubUUID)
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if lastID != lastIDForUUID {
		t.Fatalf("last update ID from GetLastUpdateID = %d and GetLastUpdateIDForUUID(store.hubUUID) = %d differ", lastID, lastIDForUUID)
	}

	// all the updates
	updList, err := store.GetUpdatesAfter(-1, -1)
	if err != nil {
		t.Fatalf("failed to fetch updates: %v", err)
	}
	for i, _ := range updList {
		updList[i].StartTime = updList[i].StartTime.UTC() // this would normally be handled by the protocol encoder
	}

	if !reflect.DeepEqual(expected, updList) {
		t.Fatalf("failed to fetch updates\nactual: %v\nexpected: %v\n", updList, expected)
	}

	// first 2
	updList, err = store.GetUpdatesAfter(0, 2)
	if err != nil {
		t.Fatalf("failed to fetch updates: %v", err)
	}
	for i, _ := range updList {
		updList[i].StartTime = updList[i].StartTime.UTC() // this would normally be handled by the protocol encoder
	}
	if len(updList) != 2 {
		t.Fatalf("failed to fetch updates: got %d updates, expected 1", len(updList))
	}
	if !reflect.DeepEqual(expected[:2], updList) {
		t.Fatalf("failed to fetch updates\nactual: %v\nexpected: %v\n", updList, expected[:2])
	}

	// 100000 after id 2
	updList, err = store.GetUpdatesAfter(2, 100000)
	if err != nil {
		t.Fatalf("failed to fetch updates: %v", err)
	}
	for i, _ := range updList {
		updList[i].StartTime = updList[i].StartTime.UTC() // this would normally be handled by the protocol encoder
	}

	if !reflect.DeepEqual(expected[2:], updList) {
		t.Fatalf("failed to fetch updates\nactual: %v\nexpected: %v\n", updList, expected[2:])
	}

	// all after id 3 -> should return nothing
	updList, err = store.GetUpdatesAfter(3, -1)
	if err != nil {
		t.Fatalf("failed to fetch updates: %v", err)
	}
	if len(updList) != 0 {
		t.Fatalf("failed to fetch updates: got %d updates, expected 0", len(updList))
	}
}

func TestForwardedUpdates(t *testing.T) {
	// prepare a new store
	os.Remove(testBoltPath)
	store, err := NewStore(StoreConfig{testBoltPath, false})
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	defer store.Close()

	// generate/append some local updates
	data, _, _ := generateTestData(10)
	if err := store.AppendMany(data); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	myLastID := int64(10)
	forwardedHub := "05defdfa-e7d1-4ca8-8b5c-02abb0088d29"

	// check if there are no updates for this hub in store
	lastID, err := store.GetLastUpdateIDForUUID(forwardedHub)
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if lastID != 0 {
		t.Fatalf("failed to get last update ID: %d, expected 0", lastID)
	}

	// get list of all hubs
	hubs, err := store.GetHubs()
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if len(hubs) != 1 {
		t.Fatalf("failed to get hub UUIDs: got %d hubs, expected 1", len(hubs))
	}
	if hubs[0] != store.hubUUID {
		t.Fatalf("fist hub should be the own stores UUID but is: %s", hubs[0])
	}

	// append 3 forwarded data-updates
	upd := updateData
	upd.StartTime = time.Date(2014, time.August, 24, 14, 35, 33, 847000000, time.UTC)
	upd.Data.Clients = clientsData

	expected := []*UpdateFull{}
	for i := int64(0); i < 3; i = i + 1 {
		in := &UpdateFull{Header{0, "", -1, "", -1}, sourceData, upd}
		in.SourceHubUUID = forwardedHub
		in.SourceHubUpdateID = 3 - i // out of order
		if err = store.Append(in); err != nil {
			t.Fatalf("unexpected error: %v", err)
		}
		myLastID = myLastID + 1
		in.ForwardHubUUID = store.GetHubUUID()
		in.ForwardHubUpdateID = myLastID
		expected = append(expected, in)
		upd.StartTime = upd.StartTime.Add(time.Duration(upd.Duration) * time.Millisecond)
	}

	out, err := store.GetUpdatesAfter(10, 3)
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	for i, _ := range out {
		out[i].StartTime = out[i].StartTime.UTC() // this would normally be handled by the protocol encoder
	}
	if !reflect.DeepEqual(out, expected) {
		t.Fatalf("failed to fetch source\nactual: %v\nexpected: %v\n", out, expected)
	}

	// check if the last update for this hub is 3
	lastID, err = store.GetLastUpdateIDForUUID(forwardedHub)
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if lastID != 3 {
		t.Fatalf("failed to get last update ID: got %d updates, expected 3", lastID)
	}

	// get list of all hubs
	hubs, err = store.GetHubs()
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if len(hubs) != 2 {
		t.Fatalf("failed to get hub UUIDs: got %d hubs, expected 2", len(hubs))
	}
	if hubs[0] != store.hubUUID {
		t.Fatalf("fist hub should be the own stores UUID but is: %s", hubs[0])
	}
	if hubs[1] != forwardedHub {
		t.Fatalf("second hub UUID is wrong: %s, expected: %s", hubs[1], forwardedHub)
	}

	// check if the last update is now 13
	lastID, _ = store.GetLastUpdateID()
	if lastID != 13 {
		t.Fatalf("failed to get last update ID: got %d updates, expected 3", lastID)
	}
}

func checkForwardedUpdates2(t *testing.T, src1Store, src2Store, fwdStore, finalStore *Store, fwdSrc1ID, fwdSrc2ID, finalSrc1ID, finalSrc2ID, finalFwdID int64) {
	lastID, err := fwdStore.GetLastUpdateIDForUUID(src1Store.GetHubUUID())
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if lastID != fwdSrc1ID {
		t.Fatalf("failed to get last update ID: %d, expected %d", lastID, fwdSrc1ID)
	}
	lastID, err = fwdStore.GetLastUpdateIDForUUID(src2Store.GetHubUUID())
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if lastID != fwdSrc2ID {
		t.Fatalf("failed to get last update ID: %d, expected %d", lastID, fwdSrc2ID)
	}
	lastID, err = finalStore.GetLastUpdateIDForUUID(src1Store.GetHubUUID())
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if lastID != finalSrc1ID {
		t.Fatalf("failed to get last update ID: %d, expected %d", lastID, finalSrc1ID)
	}
	lastID, err = finalStore.GetLastUpdateIDForUUID(src2Store.GetHubUUID())
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if lastID != finalSrc2ID {
		t.Fatalf("failed to get last update ID: %d, expected %d", lastID, finalSrc2ID)
	}
	lastID, err = finalStore.GetLastUpdateIDForUUID(fwdStore.GetHubUUID())
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if lastID != finalFwdID {
		t.Fatalf("failed to get last update ID: %d, expected %d", lastID, finalFwdID)
	}
}

func TestForwardedUpdates2(t *testing.T) {
	// prepare 4 new stores
	os.Remove(testBoltPath)
	src1Store, err := NewStore(StoreConfig{testBoltPath, false})
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	defer src1Store.Close()

	os.Remove(testBoltPath2)
	src2Store, err := NewStore(StoreConfig{testBoltPath2, false})
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	defer src2Store.Close()

	os.Remove(testBoltPathFwd)
	fwdStore, err := NewStore(StoreConfig{testBoltPathFwd, false})
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	defer fwdStore.Close()

	os.Remove(testBoltPathFinal)
	finalStore, err := NewStore(StoreConfig{testBoltPathFinal, false})
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	defer finalStore.Close()

	// generate/append some updates to src
	data, _, _ := generateTestData(10)
	if err := src1Store.AppendMany(data); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	data, _, _ = generateTestData(7)
	if err := src2Store.AppendMany(data); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}

	// check last updates so far
	checkForwardedUpdates2(t, src1Store, src2Store, fwdStore, finalStore, 0, 0, 0, 0, 0)

	// forward 5 updates from src1 to fwd
	if data, err = src1Store.GetUpdatesAfter(0, 5); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if err := fwdStore.AppendMany(data); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}

	// check last updates so far
	checkForwardedUpdates2(t, src1Store, src2Store, fwdStore, finalStore, 5, 0, 0, 0, 0)

	// forward 3 updates from src2 to fwd
	if data, err = src2Store.GetUpdatesAfter(0, 3); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if err := fwdStore.AppendMany(data); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}

	// check last updates so far
	checkForwardedUpdates2(t, src1Store, src2Store, fwdStore, finalStore, 5, 3, 0, 0, 0)

	// forward 7 updates from fwd to final
	if data, err = fwdStore.GetUpdatesAfter(0, 7); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if err := finalStore.AppendMany(data); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}

	// check last updates so far
	checkForwardedUpdates2(t, src1Store, src2Store, fwdStore, finalStore, 5, 3, 5, 2, 7)

	// forward remaining updates from src1 and src2 to fwd
	if data, err = src1Store.GetUpdatesAfter(5, -1); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if err := fwdStore.AppendMany(data); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if data, err = src2Store.GetUpdatesAfter(3, -1); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if err := fwdStore.AppendMany(data); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}

	// check last updates so far
	checkForwardedUpdates2(t, src1Store, src2Store, fwdStore, finalStore, 10, 7, 5, 2, 7)

	// forward remainging from fwd to final
	if data, err = fwdStore.GetUpdatesAfter(7, -1); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	if err := finalStore.AppendMany(data); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}

	// check last updates so far
	checkForwardedUpdates2(t, src1Store, src2Store, fwdStore, finalStore, 10, 7, 10, 7, 17)
}

func TestGetSources(t *testing.T) {
	// prepare a new store
	os.Remove(testBoltPath)
	store, err := NewStore(StoreConfig{testBoltPath, false})
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	defer store.Close()

	// check if source list is empty
	srcList, err := store.GetSources()
	if err != nil {
		t.Fatalf("fetching sources failed: %v", err)
	}
	if len(srcList) != 0 {
		t.Fatalf("wrong number of sources: %d, expected 0", len(srcList))
	}

	// generate/append some data
	data, numSrcs, _ := generateTestData(-1)
	if err := store.AppendMany(data); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}

	// fetch all the sources
	srcList, err = store.GetSources()
	if err != nil {
		t.Fatalf("fetching sources failed: %v", err)
	}
	if len(srcList) != numSrcs {
		t.Fatalf("wrong number of sources: %d, expected %d", len(srcList), numSrcs)
	}
	// the result will be orderd using the slug so doing a DeepEqual doesn't work here
}

func TestGetClients(t *testing.T) {
	// prepare a new store
	os.Remove(testBoltPath)
	store, err := NewStore(StoreConfig{testBoltPath, false})
	if err != nil {
		t.Fatalf("unexpected error: %v", err)
	}
	defer store.Close()

	// check if client list is empty
	clientList, err := store.GetClients()
	if err != nil {
		t.Fatalf("fetching clients failed: %v", err)
	}
	if len(clientList) != 0 {
		t.Fatalf("wrong number of clients: %d, expected 0", len(clientList))
	}

	// generate/append some data
	data, _, numClients := generateTestData(-1)
	if err := store.AppendMany(data); err != nil {
		t.Fatalf("unexpected error: %v", err)
	}

	// fetch all the clients
	clientList, err = store.GetClients()
	if err != nil {
		t.Fatalf("fetching clients failed: %v", err)
	}
	if len(clientList) != numClients {
		t.Fatalf("wrong number of clients: %d, expected %d", len(clientList), numClients)
	}
	// the result will be orderd using the slug so doing a DeepEqual doesn't work here
}

//
// Benchmarking
//

func BenchmarkAppendMany(b *testing.B) {
	os.Remove(testBoltPath)
	store, err := NewStore(StoreConfig{testBoltPath, false})
	if err != nil {
		b.Fatalf("unexpected error: %v", err)
	}
	defer store.Close()
	data, _, _ := generateTestData(b.N)

	b.ResetTimer()

	if err := store.AppendMany(data); err != nil {
		b.Fatalf("unexpected error: %v", err)
	}
}

func BenchmarkGetUpdatesAfter(b *testing.B) {
	os.Remove(testBoltPath)
	store, err := NewStore(StoreConfig{testBoltPath, false})
	if err != nil {
		b.Fatalf("unexpected error: %v", err)
	}
	defer store.Close()
	data, _, _ := generateTestData(b.N)
	if err := store.AppendMany(data); err != nil {
		b.Fatalf("unexpected error: %v", err)
	}

	b.ResetTimer()

	latestID := int64(-1)
	for {
		updates, err := store.GetUpdatesAfter(latestID, -1)
		if err != nil {
			b.Fatalf("failed to retrieve: %v", err)
		}
		if len(updates) == 0 {
			break
		}
		latestID = findMaxID(updates)
	}
}
