//
// sfive
//
// sfive - spreadspace streaming statistics suite is a generic
// statistic collection tool for streaming server infrastuctures.
// The system collects and stores meta data like number of views
// and throughput from a number of streaming servers and stores
// it in a global data store.
// The data acquisition is designed to be generic and extensible in
// order to support different streaming software.
// sfive also contains tools and applications to filter and visualize
// live and recorded data.
//
//
// Copyright (C) 2014-2017 Christian Pointner <equinox@spreadspace.org>
//                         Markus Grüneis <gimpf@gimpf.org>
//
// This file is part of sfive.
//
// sfive is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License version 3
// as published by the Free Software Foundation.
//
// sfive is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with sfive. If not, see <http://www.gnu.org/licenses/>.
//

package sfive

import (
	"errors"
	"time"
)

type Stream struct {
	ContentID string `json:"content"`
	Format    string `json:"format"`
	Quality   string `json:"quality"`
}

type Source struct {
	Hostname string   `json:"hostname"`
	Stream   Stream   `json:"stream"`
	Tags     []string `json:"tags,omitempty"`
}

type GeoInfo struct {
	CountryName  string  `json:"country,omitempty"`
	CountryCode2 string  `json:"country-code2,omitempty"`
	RegionName   string  `json:"region,omitempty"`
	RegionCode   string  `json:"region-code,omitempty"`
	CityName     string  `json:"city,omitempty"`
	Latitude     float64 `json:"latitude,omitempty"`
	Longitude    float64 `json:"longitude,omitempty"`
}

type Client struct {
	IP        string `json:"ip"`
	Port      uint   `json:"port,omitempty"`
	UserAgent string `json:"user-agent,omitempty"`
	BytesSent uint   `json:"bytes-sent,omitempty"`
	GeoInfo
}

type UpdateData struct {
	ClientCount   uint     `json:"client-count"`
	BytesReceived uint     `json:"bytes-received,omitempty"`
	BytesSent     uint     `json:"bytes-sent"`
	Clients       []Client `json:"clients,omitempty"`
}

type Update struct {
	StartTime time.Time  `json:"start-time"`
	Duration  int64      `json:"duration-ms"`
	Data      UpdateData `json:"data"`
}

type Header struct {
	Version            uint   `json:"version,omitempty"` // omitempty is needed for data only messages and for REST API
	SourceHubUUID      string `json:"SourceHubUuid,omitempty"`
	SourceHubUpdateID  int64  `json:"SourceHubUpdateId,omitempty"`
	ForwardHubUUID     string `json:"ForwardHubUuid,omitempty"`
	ForwardHubUpdateID int64  `json:"ForwardHubUpdateId,omitempty"`
}

type UpdateFull struct {
	Header
	Source
	Update
}

func (uf *UpdateFull) checkSourceAndTime() error {
	if uf.Source.Hostname == "" {
		return errors.New("empty hostname is not allowed")
	}
	if uf.Source.Stream.ContentID == "" {
		return errors.New("empty content-id is not allowed")
	}
	if uf.Source.Stream.Format == "" {
		return errors.New("empty format is not allowed")
	}
	if uf.Source.Stream.Quality == "" {
		return errors.New("empty quality is not allowed")
	}
	if uf.StartTime.IsZero() {
		return errors.New("empty start-time is not allowed")
	}
	if uf.Duration < 1 {
		return errors.New("duration must be > 0")
	}
	return nil
}

func (uf *UpdateFull) CopyFromSource(src *Source) {
	uf.Hostname = src.Hostname
	uf.Stream = src.Stream
	uf.Tags = src.Tags
}

func (uf *UpdateFull) CopyFromUpdate(u *Update) {
	uf.StartTime = u.StartTime
	uf.Duration = u.Duration
	uf.Data = u.Data
}

// web

type WebErrorResponse struct {
	Error string `json:"error,omitempty"`
}

type WebHealthzResponse struct {
	Status  string `json:"status"`
	HubUUID string `json:"hub-uuid"`
}

type WebHubsResponse struct {
	Hubs []string `json:"hubs"`
}

type WebSourcesResponse struct {
	Sources []*Source `json:"sources"`
}

type WebClientsResponse struct {
	Clients []*Client `json:"clients"`
}

type WebUpdatesGetResponse struct {
	Updates []*UpdateFull `json:"updates"`
}

type WebUpdatesPostResponse struct {
	NumUpdates int64 `json:"num-updates"`
}

type WebLastUpdateIDResponse struct {
	HubUUID      string `json:"hub-uuid"`
	LastUpdateID int64  `json:"lastupdate"`
}
